import os
import time
import traceback
from threading import Thread
from sshkeyboard import listen_keyboard, stop_listening
from rich.columns import Columns
from rich.panel import Panel
from rich.text import Text
from rich.live import Live
from rich.table import Table
from rich.layout import Layout
from rich.align import Align
from dry_pipe.monitoring import PipelineMetricsTable, fetch_task_group_metrics
from dry_pipe.pipeline import PipelineInstance
from dry_pipe.task_state import TaskState


class CliScreen:

    def __init__(self, pipeline_instance_dir, is_monitor_mode):
        self.is_monitor_mode = is_monitor_mode
        self.pipeline_instance_dir = pipeline_instance_dir
        self.pipeline_hints = PipelineInstance.load_hints(pipeline_instance_dir)
        self.quit = False
        self.screen = 'summary'
        self.failed_task_states = []
        self.failed_task_states_iter_count = 0
        self.selected_failed_task = None
        self.prompt = None
        self.error_msg = []

        class ShallowPipelineInstance:
            def __init__(self):
                self.pipeline_instance_dir = pipeline_instance_dir

        self.pipeline_instance = ShallowPipelineInstance()

    def fetch_failed_task_states(self):
        if self.failed_task_states_iter_count % 3 == 0:
            ts = [
                task_state for task_state in TaskState.failed_task_states(self.pipeline_instance)
            ]
            self.failed_task_states = sorted(ts, key=lambda s: s.task_key)
            if self.selected_failed_task is None and len(self.failed_task_states) > 0:
                self._set_selected_failed_task(0)

        self.failed_task_states_iter_count += 1

    def _set_selected_failed_task(self, idx):
        self.selected_failed_task = self.failed_task_states[idx]

    def _selected_failed_task_index(self):
        if self.selected_failed_task is None:
            return None

        for i in range(0, len(self.failed_task_states)):
            if self.failed_task_states[i].task_key == self.selected_failed_task.task_key:
                return i

    def restart_failed_task(self):
        self.selected_failed_task.transition_to_queued()

    def task_up(self):
        c = len(self.failed_task_states)
        if c == 0:
            return
        idx = self._selected_failed_task_index()
        if idx is None or idx == 0:
            self._set_selected_failed_task(0)
        else:
            idx -= 1
            if idx < c:
                self._set_selected_failed_task(idx)

    def task_down(self):
        c = len(self.failed_task_states)
        if c == 0:
            return
        idx = self._selected_failed_task_index()
        if idx is None:
            self._set_selected_failed_task(0)
        else:
            idx += 1
            if idx < c:
                self._set_selected_failed_task(idx)
            else:
                self._set_selected_failed_task(c - 1)


    def press(self, key):
        try:
            if key == "q":
                self.quit = True
                stop_listening()
            elif key == "f":
                self.cancel_prompt()
                self.screen = "errors"
            elif key == "s":
                self.cancel_prompt()
                self.screen = "summary"
            elif key == "r":
                if self.selected_failed_task is not None:
                    self.set_prompt_restart_task()
            elif key == "n":
                self.cancel_prompt()
            elif key == "y":
                if self.is_prompt_restart_task():
                    self.restart_failed_task()
            elif key == "up":
                self.cancel_prompt()
                self.task_up()
            elif key == "down":
                self.cancel_prompt()
                self.task_down()
        except Exception:
            self.error_msg.append(traceback.format_exc())
            stop_listening()

    def is_prompt_restart_task(self):
        return self.prompt == "restart-task"

    def set_prompt_restart_task(self):
        self.prompt = "restart-task"

    def cancel_prompt(self):
        self.prompt = None

    def start_and_wait(self):
        def refresher():
            with Live(refresh_per_second=2) as live:
                try:
                    while not self.quit:
                        self.update_screen(live)
                        time.sleep(1)
                except Exception:
                    self.error_msg.append(traceback.format_exc())
                    stop_listening()

        Thread(target=refresher).start()

        listen_keyboard(on_press=lambda key: self.press(key))

        if len(self.error_msg) > 0:
            return self.error_msg[0]
        else:
            return None

    def _errors_screen(self):

        failed_cnt = len(self.failed_task_states)

        layout = self._main_screen()

        if failed_cnt == 0:
            layout["header3"].update("Failed Tasks")
            layout["body"].update("...there are no failed tasks")
            self.cancel_prompt()
            return layout

        def get_header_text():

            if not self.is_prompt_restart_task():
                t = Text()
                t.append(f"Task({self.selected_failed_task.task_key}) -> ", style="bold magenta")
                err_file = os.path.join(self.selected_failed_task.control_dir(), "err.log")
                t.append(f"tail -50 {err_file}\n", style="bold magenta")
                t.append(f"key r to restart", style="bold gree1")
                return t
            else:
                t = Text()
                t.append(f"Restart task {self.selected_failed_task.task_key} ? y/N/a (a=all failed tasks)", style="bold green1")
                return t

        def get_tail_text():
            t = Text()
            t.append(self.selected_failed_task.tail_err_if_failed(50), style="red")
            return t

        layout["header3"].update(get_header_text())
        layout["body"].update(get_tail_text())

        return layout

    def _main_screen(self):
        layout = Layout()
        layout.split_column(
            Layout(name="header1", size=1),
            Layout(name="header2", size=1),
            Layout(name="header3", size=1),
            Layout(name="body", ratio=10),
            Layout(name="footer", size=1)
        )

        pipeline_mod_func = self.pipeline_hints.get("pipeline")
        if self.is_monitor_mode:
            console_mode = "  [bright_yellow]console in monitor mode (no launching)[/]"
        else:
            console_mode = "  [bright_yellow]console in launcher mode[/]"

        if pipeline_mod_func is not None:
            title = f"[bold]DryPipe([green1]{pipeline_mod_func}[/], "
            title += f"[cyan1]{self.pipeline_instance.pipeline_instance_dir}[/])  {console_mode}"
        else:
            title = f"DryPipe - {console_mode}"

        help_panel = Columns([
            "command keys: "
            "[b]s[/]: summary",
            "[b]f[/]: failed tasks",
            "[b]q[/]: quit"
        ], equal=True, expand=False)

        layout["header1"].update(Align(title, align="center"))
        layout["header2"].update(Align(help_panel, align="left"))

        return layout

    def _status_table(self):

        table = Table(
            show_header=True,
            header_style="none",
            show_lines=False,
            show_edge=False,
            expand=True
        )

        header, body, footer = PipelineMetricsTable.summarized_table_from_task_group_metrics(
            fetch_task_group_metrics(self.pipeline_instance)
        )

        for h in header:
            table.add_column(h)

        def color_row(row):
            c = 0
            row_cells = []
            for v in row:
                if v is None or v == 0:
                    row_cells.append("")
                else:
                    if c == 3:
                        row_cells.append(f"[green1]{v}[/]")
                    elif c in [6, 7, 9]:
                        row_cells.append(f"[red]{v}[/]")
                    elif c in [0, 1]:
                        row_cells.append(f"[yellow]{v}[/]")
                    else:
                        row_cells.append(f"{v}")
                c += 1
            return row_cells

        for row in body:
            table.add_row(*color_row(row))

        table.add_row(*color_row(footer))

        layout = self._main_screen()

        layout["header3"].update("Task Execution Status Summary")
        layout["body"].update(table)

        return layout

    def update_screen(self, live):
        if self.screen == 'summary':
            l = self._status_table()
        elif self.screen == "errors":
            self.fetch_failed_task_states()
            l = self._errors_screen()
        elif self.screen == "restart-task":
            l = self._errors_screen()

        live.update(Panel(l, border_style="blue"))






def bash_shebang():
    return "#!/usr/bin/env bash"

BASH_TASK_FUNCS_AND_TRAPS = """


function __check_bash_version () {
    __bash_version_fail="DryPipe requires bash version >= 4.3.x, current version is: $BASH_VERSION"
    
    if [ -z "${BASH_VERSINFO}" ] || [ -z "${BASH_VERSINFO[0]}" ]; then
        echo $__bash_version_fail >&2         
        __transition_to_failed "" ""
    fi
                
    if [[ ${BASH_VERSINFO[0]} -lt 4  || ( ${BASH_VERSINFO[0]} -eq 4  && ${BASH_VERSINFO[1]} -lt 3 ) ]]; then 
        echo $__bash_version_fail >&2
        __transition_to_failed        
    fi
}    
                
function __read_task_state () {

    local count_for_sanity=$(find $__control_dir/state.* | wc -l)

    if [[ $count_for_sanity -ne 1 ]]; then
        echo "CORRUPTED_CONTROL_DIR: exactly one state file is required, found $count_for_sanity in $__control_dir" >&2
        exit 1
    fi

    local __state_file=$(ls $__control_dir/state.* | tr -d '[:space:]')
    local __state_file_name=$(basename $__state_file)
    export __state_file_abs_path="$__control_dir/$__state_file_name"
    local state_parts=(${__state_file_name//./ })

    export __state_name=${state_parts[1]}

    export __next_step_number=$(( ${state_parts[2]} + 0 ))
    
   if [[ $__state_name != "launched" && $__state_name != "scheduled" && $__state_name != "failed" && $__state_name != "timed-out" ]]; then

        if [[ $__arg_1 == "--force" ]]; then
            echo "force launching task $__task_key"
        else
            echo "INVALID_STATE_FOR_LAUNCHING: task state must be either launched or scheduled, state file: $__state_file" >&2
            echo "If you want to launch task script manualy, use --force as first argument" >&2
            exit 1
        fi
    fi
    
}

function __invoke_step () {    
    if [[ $__is_slurm == "True" ]]; then
        $__step_command &
        wait $!
    else 
        $__step_command
    fi    
}

function __touch_history () {
    touch $__control_dir/step_start
}

function __append_history () {

    local __t=$(date +"%FT%T.%5N" -d @$(stat -c %.5Y $__control_dir/step_start ))
    local __step_to_record=$__next_step_number
    
    if [[ $__state_name == "step-completed" ]]; then        
        __step_to_record=$((__next_step_number-1))
    elif [[ $__state_name == "completed-unsigned" ]]; then
        __step_to_record=""
    fi
    
    echo "$__state_name\t$__t\t$__step_to_record" >> "$__control_dir/history.tsv"                 
}

function __transition_to_next_state () {

    local next_state_file_abs_path="$__control_dir/state.$1.${__next_step_number}"

    mv $__state_file_abs_path $next_state_file_abs_path

    __state_name=$1
    __state_file_abs_path=$next_state_file_abs_path
            
    # touch $__pipeline_instance_dir/.drypipe/state.*  || true        
}

function __transition_to_failed () {
     
    __append_history
    __touch_history
    
    echo "task failed"
    local parent_lineno="$1"
    echo "Failed at line ${parent_lineno}"  >&2
    __transition_to_next_state "failed"
    exit 1
}

function __transition_to_timed_out () {

    __append_history
    __touch_history
        
    echo "task timed out"
    echo "task timed out"  >&2
    __transition_to_next_state "timed-out"    
    exit 1
}

function __transition_to_step_started () {

    __append_history
    __touch_history
    
    __transition_to_next_state "step-started"
}

function __transition_to_step_completed () {

    __append_history
    __touch_history
    
    __next_step_number=$((__next_step_number+1))
    __transition_to_next_state "step-completed"
}

function __transition_to_completed () {

    __append_history
    __touch_history    

    mv "$__state_file_abs_path" "$__control_dir/state.completed-unsigned"

    rm -f $__control_dir/out.sig
    
    __state_name="completed-unsigned"
    __touch_history
    __append_history    
}

function __container_image_path () {

    if [ -f "$1" ]; then
        echo "$1"    
    elif [ -f "$__containers_dir/$1" ]; then
        echo "$__containers_dir/$1"
    else 
        echo "container $1 not found" >&2
        echo "defined containers_dir=$__containers_dir" >&2
        exit 1;       
    fi        
}

function __add_binding_for_singularity_if_required () {

    if [[ ! $____singularity_bindings_done == "True" ]]; then
        
        local __root_dir_of_script=$(echo "$__pipeline_code_dir" | cut -d "/" -f2)
        local __fs_type=$(stat -f -L -c %T "/$__root_dir_of_script")
        if [[ $__fs_type == "autofs" || $__fs_type == "nfs" || $__fs_type == "zfs" ]]; then
    
            local __bind="/$__root_dir_of_script:/$__root_dir_of_script"
            
            if [ -z ${SINGULARITY_BIND+x} ]; then 
              export SINGULARITY_BIND=$__bind         
            else 
              export SINGULARITY_BIND=$__bind,$SINGULARITY_BIND
            fi            
        fi
    
        if [[ $__is_slurm == "True" ]]; then
            local __root_of_scratch=$(echo "$__scratch_dir" | cut -d "/" -f2)            
            export SINGULARITY_BIND=$SINGULARITY_BIND,/$__root_of_scratch:/$__root_of_scratch
        fi    
        export ____singularity_bindings_done=True
    fi    
}    

"""

__BASH_SIGN_FILES = """
function __sign_file () {
  fn=$(basename $1)
  if [ ! -f "$1" ]; then
     echo "warning: task $__task_key didn't produce output file $fn as specified"
  else
     fn="$__control_dir/$__sig_dir/$fn.sig"
%s
     sha1sum $1 > $fn
%s
  fi
}

function __sign_files () {

    for f in $__fileset_to_sign; do
        __sign_file $f
    done

    for f in $(echo $__file_list_to_sign | sed "s/,/ /g"); do
        __sign_file $f
    done
}

"""

BASH_SIGN_FILES = __BASH_SIGN_FILES % ("", "")

BASH_SIGN_FILES_IF_NEWER = __BASH_SIGN_FILES % (
    "    if [[ $1 -nt $fn ]]; then",
    "    fi"
)



import glob
import inspect
import json
import os
import subprocess
import sys
import getpass
import logging.config

import psutil
from psutil import AccessDenied

from dry_pipe import env_variables

LOCAL_PROCESS_IDENTIFIER_VAR = "____DRY_PIPE_TASK"

log_conf_file = env_variables.log_config()


def DEBUG_CONFIG(main_level, drypype_level, root_level):
    return {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'formatter': 'standard',
                'class': 'logging.StreamHandler',
                'stream': 'ext://sys.stdout',
            },
        },
        'loggers': {

            'zoz': {  # root logger
                'handlers': ['console'],
                'level': root_level,
                'propagate': False
            },
            'drypipe.*': {
                'handlers': ['console'],
                'level': drypype_level,
                'propagate': False
            },
            '__main__': {
                'handlers': ['console'],
                'level': main_level,
                'propagate': False
            }
        }
    }


def setup_debug_log_config(drypype_level="INFO", root_level="INFO", main_level="INFO", extra_loggers={}):
    print("DEBUG CONFIG")

    conf = DEBUG_CONFIG(main_level, drypype_level, root_level)

    conf["loggers"] = {
        ** conf["loggers"],
        ** extra_loggers
    }

    logging.config.dictConfig(conf)

if log_conf_file is not None:
    import yaml
    if log_conf_file == "DEBUG_CONFIG":
        setup_debug_log_config()
    else:
        with open(log_conf_file) as f:

            print(f"log config: {log_conf_file}")
            config = yaml.load(f, Loader=yaml.FullLoader)
            logging.config.dictConfig(config)


logger = logging.getLogger(__name__)


class ValidationError(Exception):
    def __init__(self, message, code=None):
        super(ValidationError, self).__init__(message)
        self.code = code

    consumes_has_bad_positional_arg_val = 0

    consumes_has_bad_positional_arg = 1

    consumes_has_invalid_kwarg_type = 2

    produces_cant_take_positional_args = 3

    produces_only_takes_files = 4

    call_has_bad_arg = 5


class Val:

    def __init__(self, value):
        t = type(value)
        if t == int or t == str or t == float:
            self.value = value
        elif t == type(int):
            raise ValidationError(f"invalid val type dsl.val(int), you probably want to use 'dsl.var(int)'")
        elif t == t == type(str):
            raise ValidationError(f"invalid val type dsl.var(str), you probably want to use 'dsl.var(str)'")
        elif type(float):
            raise ValidationError(f"invalid val type, dsl.var(float) you probably want to use 'dsl.var(float)'")
        else:
            raise ValidationError(f"invalid val type: {type(value)}, must be int, string, float")

    def to_hash(self):
        return str(self.value)

    def serialized_value(self):

        if type(self.value) == str:
            return f'"{self.value}"'

        return f"{self.value}"


class IncompleteVar:

    def __init__(self, type=str, may_be_none=True):
        self.type = type
        self.may_be_none = may_be_none


class OutputVar:

    def __init__(self, type, may_be_none, name, producing_task):

        supported = [str, int, float]

        if type not in supported:
            raise ValidationError(f"var {name} can't be of type {type} supported types are: {','.join(supported)}")

        self.name = name
        self.type = type
        self.producing_task = producing_task
        self.may_be_none = may_be_none

    def format_for_python(self, v):

        if self.type == int:
            return str(v)
        elif self.type == str:
            return f'"{v}"'
        elif self.type == float:
            return str(v)

        raise Exception(f"unsupported type {self.type}")

    def unformat_for_python(self, v):

        if self.type == int:
            return int(v)
        elif self.type == str:
            s = v.rstrip('\"\'')
            s = s.lstrip('\"\'')
            return s
        elif self.type == float:
            return float(v)

        raise Exception(f"unsupported type {self.type}")

    def input_var(self, var_name_in_consuming_task):
        return InputVar(self, var_name_in_consuming_task)

    def fetch(self):
        varz = self.producing_task.out_vars_by_name()
        v = varz.get(self.name)
        return self.unformat_for_python(v)


class InputVar:

    def __init__(self, output_var, var_name_in_consuming_task):
        self.output_var = output_var
        self.var_name_in_consuming_task = var_name_in_consuming_task

    def write(self, string_value):
        return string_value


class Executor:

    def key(self):
        return None

    def is_remote(self):
        return False

class Local(Executor):

    def count_running_tasks(self):

        current_user = getpass.getuser()

        c = 0

        for it in psutil.process_iter(['pid', 'exe', 'username', 'environ']):

            if it.username() != current_user:
                continue
            try:

                if not it.exe().endswith("/bash"):
                    continue

                v = it.environ().get(LOCAL_PROCESS_IDENTIFIER_VAR)

                if v is not None: # and v.startswith("____"):
                    c += 1

            except AccessDenied:
                pass

        return c

    def kill_task(self, task_key):

        current_user = getpass.getuser()

        for it in psutil.process_iter(['pid', 'exe', 'username', 'environ']):

            if it.username() != current_user:
                continue
            try:

                if not it.exe().endswith("/bash"):
                    continue

                v = it.environ().get(LOCAL_PROCESS_IDENTIFIER_VAR)

                if v is not None:
                    if v == f"____{task_key}":
                        it.kill()

            except AccessDenied:
                pass

        #pids = list(get_pids())
        #print(f"pids: {pids}")
        #for pid in pids:
        #    os.kill(pid. signal.SIGKILL)

    def __init__(self, before_execute_bash):
        self.before_execute_bash = before_execute_bash
        self.cpu_count = len(psutil.Process().cpu_affinity())

    def has_cpu_capacity_to_launch(self):

        return self.count_running_tasks() <= self.cpu_count

    fail_silently_for_test = False

    """
        experimental, for scenario:
        + launcher runs in a LXC container
        + launcher runs tasks with singularity,launch on the physical host of the LXC container  
    """
    def __command_on_sibling_container(self, task, user, back_ground):
        return [
            f"ssh {user}@127.0.0.1 'nohup {task.v_abs_script_file()} " +
            f"  >{task.v_abs_out_log()} 2>{task.v_abs_err_log()} {back_ground}'"
        ]

    def execute(self, task, touch_pid_file_func, wait_for_completion=False, fail_silently=False):
        b4_command = ""
        if self.before_execute_bash is not None:
            b4_command = f"{self.before_execute_bash} &&"

        back_ground = "&"

        if wait_for_completion:
            back_ground = ""

        with subprocess.Popen(
                [f"nohup bash -c '{b4_command} . {task.v_abs_script_file()} " +
                 f"  >{task.v_abs_out_log()} 2>{task.v_abs_err_log()}' {back_ground}"],
                shell=True,
                text=True,
                env={
                    **dict([
                        (LOCAL_PROCESS_IDENTIFIER_VAR, f"____{task.key}")
                    ]),
                    **os.environ
                },
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
        ) as p:

            p.wait()

            if p.returncode != 0:

                # If we were in async mode, we wouldn't see the error...
                if wait_for_completion:
                    if fail_silently:
                        return

                    if Local.fail_silently_for_test:
                        return

                    task_state = task.get_state()
                    err_if_failed = task_state.tail_err_if_failed(5)

                    if err_if_failed is not None:
                        raise Exception(
                            f"task {task.key} failed \n{err_if_failed}\n further details in {task.v_abs_err_log()}"
                        )
                    else:
                        return

                raise Exception(
                    f"the command for {task} returned non zero result, " +
                    f"see logs at {task.v_abs_out_log()} and {task.v_abs_err_log()}.\n",
                    f"invocation script: {task.v_abs_script_file()}"
                )

            touch_pid_file_func(p.pid)


class Slurm(Executor):

    def __init__(self, account, sbatch_options):

        if account is None:
            raise Exception(f"account account can't be None")

        self.account = account
        self.sbatch_options = sbatch_options

    def execute(self, task, touch_pid_file_func, wait_for_completion=False, fail_silently=False):

        env = {
            **os.environ
        }

        if wait_for_completion:
            cmd = [f"bash -c 'export SBATCH_WAIT=True && {task.v_abs_sbatch_launch_script()}'"]
        else:
            cmd = [task.v_abs_sbatch_launch_script()]

        with subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            env=env,
            text=True
        ) as p:
            p.wait()

            if p.returncode != 0:
                err = p.stderr.read().strip()
                raise Exception(f"call to sbatch failed: {cmd}\n {err}")


class IndeterminateFile:

    def __init__(self, file_path, manage_signature):
        self.file_path = file_path.strip()
        self.manage_signature = manage_signature

    def produced_file(self, var_name, producing_task):
        return ProducedFile(self.file_path, var_name, self.manage_signature, producing_task)

    def pre_existing_file(self):
        return PreExistingFile(self.file_path, self.manage_signature)


class FileSet:

    def __init__(self, glob_pattern):
        self.glob_pattern = glob_pattern

    def out_file_set(self, producing_task, name_in_producing_task):
        return OutFileSet(self, producing_task, name_in_producing_task)


class OutFileSet:

    def __init__(self, file_set, producing_task, name_in_producing_task):
        self.file_set = file_set
        self.producing_task = producing_task
        self.name_in_producing_task = name_in_producing_task

    def fetch(self):
        if not self.producing_task.get_state().is_completed():
            raise Exception(f"task must be complete before out value can be fetched")

        producing_task_work_dir = self.producing_task.v_abs_work_dir()

        pattern = os.path.join(producing_task_work_dir, self.file_set.glob_pattern)

        def make_path_relative_to_pipeline_instance_dir(p):
            return os.path.relpath(p, producing_task_work_dir)

        for f in glob.glob(pattern):
            f = make_path_relative_to_pipeline_instance_dir(f)
            yield ProducedFile(f, self.name_in_producing_task, True, self.producing_task)


class SubPipeline:
    def __init__(self, pipeline, task_namespance_prefix, dsl):
        self.pipeline = pipeline
        self.task_namespance_prefix = task_namespance_prefix
        self.dsl = dsl

    def with_completed_tasks(self, *args):
        args = [f"{self.task_namespance_prefix}{a}" for a in args]
        return self.dsl.with_completed_tasks(*args)


class TaskMatcher:

    def __init__(self, task_keys_glob_pattern):
        self.task_keys_glob_pattern = task_keys_glob_pattern


class PreExistingFile:

    def __init__(self, file_path, manage_signature=False):

        if file_path is None or file_path == "" or file_path == ".":
            raise ValidationError(f"invalid file {file_path}")

        self.file_path = file_path
        self.manage_signature = manage_signature

    def to_hash(self):
        return self.file_path

    def value(self):
        return self.file_path

    def absolute_path(self, task):

        if os.path.isabs(self.file_path):
            return self.file_path

        return self.file_path


class ProducedFile:

    def __init__(self, file_path, var_name, manage_signature, producing_task, is_dummy=False):

        if file_path is None or file_path == "" or file_path == ".":
            raise ValidationError(f"invalid file {file_path}")

        if var_name is None or var_name == "":
            raise ValidationError(f"can't be {var_name}")

        self.var_name = var_name
        self.file_path = file_path
        self.producing_task = producing_task
        self.manage_signature = manage_signature
        self.is_dummy = is_dummy

    def to_hash(self):
        return self.file_path

    def value(self):
        return self.file_path

    def absolute_path(self, task):

        if os.path.isabs(self.file_path):
            return self.file_path

        return os.path.join(task.work_dir(), self.file_path)

    def input_file(self, var_name_in_consuming_task):
        return InputFile(self, var_name_in_consuming_task)


class InputFile:

    def __init__(self, produced_file, var_name_in_consuming_task):
        self.produced_file = produced_file
        self.var_name_in_consuming_task = var_name_in_consuming_task

def env_from_sourcing(env_file):

    p = os.path.abspath(sys.executable)

    dump_with_python_script = f'{p} -c "import os, json; print(json.dumps(dict(os.environ)))"'

    with subprocess.Popen(
            ['/bin/bash', '-c', f". {env_file} && {dump_with_python_script}"],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    ) as pipe:
        pipe.wait()
        err = pipe.stderr.read()
        out = pipe.stdout.read()
        if pipe.returncode != 0:
            raise Exception(f"Failed sourcing env file: {env_file}\n{err}")
        return json.loads(out)


class PythonCall:

    def __init__(self, func, tests=[]):
        self.func = func
        self.signature = inspect.signature(self.func)
        self.tests = tests

    def signature_spec(self):
        raise Exception(f"implement me")

    def mod_func(self):
        mod = inspect.getmodule(self.func)
        file_name = os.path.basename(mod.__file__)
        func_name = self.func.__name__
        importable_module_name = f"{mod.__package__}.{file_name[:-3]}:{func_name}"
        return importable_module_name


class Wait:

    def __init__(self, tasks):
        self.tasks = tasks

    def is_ready(self):
        for t in self.tasks:
            if not t.has_completed():
                return False

        return True

class TaskProps:

    def __init__(self, task, props):
        self.props = props
        self.task = task

    def __getattr__(self, name):
        if name not in self.props:
            raise ValidationError(f"{self.task} has no prop {name}")
        return self.props.get(name)


def flatten(the_list):
    return [item for sublist in the_list for item in sublist]


class MessageUtils:

    @staticmethod
    def safe_utf8(s):
        try:
            return s.decode("utf8")
        except UnicodeDecodeError as ex:
            return str(s)




import glob
import importlib
import inspect
import json
import pathlib
import re
import sys
import os
import traceback
import click
from dry_pipe import DryPipe
from dry_pipe.internals import env_from_sourcing
from dry_pipe.janitors import Janitor
from dry_pipe.monitoring import fetch_task_groups_stats
from dry_pipe.pipeline import Pipeline, PipelineInstance
from dry_pipe.pipeline_state import PipelineState
from dry_pipe.task_state import NON_TERMINAL_STATES



@click.group()
@click.option('-v', '--verbose')
@click.pass_context
def cli_group(ctx, verbose):
    ctx.ensure_object(dict)
    ctx.obj['verbose'] = verbose
    ctx.obj["ctx"] = ctx


def _bad_entrypoint(ctx):

    if "entrypoint" in ctx.obj and ctx.obj["entrypoint"] == "run_task":
        click.echo("Can't launch a pipeline with cli.run_task(), use cli.run(pipeline=<your pipeline>)", err=True)
        return True

    return False

@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
def clean(ctx, instance_dir):
    if _bad_entrypoint(ctx):
        return
    pipeline_func = ctx.obj["pipeline_func"]
    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)
    # TODO: prompt
    pipeline.clean()


def _pipeline_from_pipeline_func(pipeline_func, instance_dir):

    pipeline_func_sig = inspect.signature(pipeline_func)

    nargs = len(pipeline_func_sig.parameters)

    if nargs == 0:

        if instance_dir is not None:
            raise Exception(f"when specifying instance_dir, task generator must take a dry_pipe.DryPipe.dsl() argument")

        return DryPipe.create_pipeline(pipeline_func)
    elif nargs == 1:
        #dsl = DryPipe.dsl_for(pipeline_func)
        pipeline_code_dir = DryPipe.pipeline_code_dir_for(pipeline_func)
        pipeline_instance_dir = pipeline_code_dir

        #if instance_dir is not None:
        #    dsl = dsl.with_defaults(pipeline_instance_dir=pipeline_code_dir)
        return DryPipe.create_pipeline(pipeline_func, pipeline_code_dir)
    else:
        m = inspect.getmodule(pipeline_func)
        raise Exception(f"function {pipeline_func.__name__} in {m.__file__} takes {nargs} arguments,\n"
                        f"it should take a instance created with dry_pipe.DryPipe.dsl()")


@click.command()
@click.pass_context
@click.option('--clean', is_flag=True)
@click.option('--instance-dir', type=click.Path(), default=None)
def prepare(ctx, clean, instance_dir):

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if clean:
        pipeline.clean_all()

    pipeline.init_work_dir()

    pipeline.get_state().touch()

    tasks_total = 0
    work_done = 0
    tasks_in_non_terminal_states = 0
    tasks_completed = 0

    for task in pipeline.tasks:

        tasks_total += 1

        task_state = task.get_state()

        if task_state is None:
            task.create_state_file_and_control_dir()
            work_done += 1
            task_state = task.get_state()

        if task_state.state_name in NON_TERMINAL_STATES:
            tasks_in_non_terminal_states += 1

        if task_state.is_waiting_for_deps():

            if not task.has_unsatisfied_deps():
                if task_state.is_prepared():
                    continue
                else:
                    task_state.transition_to_prepared(task)
                    task_state = task.get_state()
                    task_state.transition_to_queued(task)
                    task.prepare()
                    tasks_in_non_terminal_states += 1
                    work_done += 1

        elif task_state.is_completed():
            if task_state.action_if_exists() is None:
                tasks_completed += 1

    if tasks_total == tasks_completed:
        pipeline_state = pipeline.get_state()
        if not pipeline_state.is_completed():
            pipeline_state.transition_to_completed()


def _pipeline_from_modul_func(instance_dir, module_func_pipeline, write_hint_file_if_not_exists=True):

    mod, func_name = module_func_pipeline.split(":")

    module = importlib.import_module(mod)

    func = getattr(module, func_name, None)

    if func is None:
        raise Exception(f"function {func_name} not found in module {mod}")

    pipeline = func()

    if not isinstance(pipeline, Pipeline):
        raise Exception(f"function {func_name} in {mod} should return a Pipeline, not {type(pipeline).__name__}")

    if write_hint_file_if_not_exists:
        PipelineInstance.write_hint_file_if_not_exists(instance_dir, module_func_pipeline)

    return pipeline


def _pipeline_instance_creater(instance_dir, pipeline_spec):

    if pipeline_spec is None:
        pipeline_spec = PipelineInstance.guess_pipeline_from_hints(instance_dir)

    pipeline = _pipeline_from_modul_func(instance_dir, pipeline_spec)

    return lambda: pipeline.create_pipeline_instance(instance_dir)


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
def mon(pipeline, instance_dir):

    from dry_pipe.cli_screen import CliScreen

    if instance_dir is None:
        instance_dir = os.getcwd()

    if pipeline is None:
        pipeline = PipelineInstance.guess_pipeline_from_hints(instance_dir)
        if pipeline is None:
            click.echo("must specify pipeline, ex: -p <aModule>:<aPipelineFunction>")

    #pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline)()

    screen_state = CliScreen(instance_dir, is_monitor_mode=True)

    err_msg = screen_state.start_and_wait()

    if err_msg is not None:
        click.echo(err_msg, err=True)


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--web-mon', is_flag=True)
@click.option('--port', default=5000, help="port for --web-mon")
@click.option('--bind', default="0.0.0.0", help="bind address for --web-mon")
@click.option('--clean', is_flag=True)
@click.option('--single', type=click.STRING, default=None,
              help="launches a single task, specified by the task key, in foreground, ex.: launch --single=TASK_KEY")
@click.option('--restart-failed', is_flag=True)
@click.option('--reset-failed', is_flag=True)
@click.option('--no-confirm', is_flag=True)
def run(pipeline, instance_dir, web_mon, port, bind, clean, single, restart_failed, reset_failed, no_confirm):

    pipeline_mod_func = pipeline

    if instance_dir is None:
        instance_dir = os.getcwd()

    if PipelineState.from_pipeline_instance_dir(instance_dir, none_if_not_exists=True) is None:
        if not no_confirm:
            if not click.confirm(f"no pipeline instance exists in {instance_dir}, do you want to create one ?"):
                click.echo("no pipeline instance created, exiting.")
                return

    pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline_mod_func)()

    if clean:
        pipeline_instance.clean()

    pipeline_instance.init_work_dir()

    if single is not None:
        launch_single(pipeline_instance, single)
        return

    pipeline_instance.regen_tasks_if_stale(force=True)

    for task in pipeline_instance.tasks:
        task_state = task.get_state()
        if task_state is not None:

            if task_state.is_failed() and not task.has_unsatisfied_deps():
                task.prepare()

                if restart_failed:
                    task.re_queue()

    janitor = Janitor(
        pipeline_instance=pipeline_instance,
        min_sleep=0,
        max_sleep=5
    )

    janitor.start(stay_alive_when_no_more_work=True)
    janitor.start_remote_janitors()

    if web_mon:
        from dry_pipe.websocket_server import WebsocketServer
        from dry_pipe.server import PipelineUIServer
        server = PipelineUIServer.init_single_pipeline_instance(pipeline_instance.get_state())
        server.start()

        WebsocketServer.start(bind, port)
        # Stopping threads break CTRL+C, let the threads die with the process exit
        #server.stop()
        #janitor.do_shutdown()
    else:
        from dry_pipe.cli_screen import CliScreen
        rich_screen = CliScreen(pipeline_instance.pipeline_instance_dir, is_monitor_mode=False)
        err_msg = rich_screen.start_and_wait()
        if err_msg is not None:
            click.echo(err_msg, err=True)

    os._exit(0)


def launch_single(pipeline, single):

    task = pipeline.tasks.get(single)

    if task is None:
        raise Exception(f"pipeline has no task with key '{single}'")

    task.re_queue()
    task.prepare()
    task_state = task.get_state()
    task_state.transition_to_launched(task.task_conf.create_executer(), task, wait_for_completion=True)
    task_state = task.get_state()
    task_state.transition_to_completed(task)

    print(f"Task {single} completed.")


@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--single', type=click.STRING, default=None,
              help="apply to single task, specified by the task key, in foreground, ex.: launch --single=TASK_KEY")
def requeue(ctx, single, instance_dir):

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if single is None:
        raise Exception(f"unimplemented.")

    requeue_single(pipeline, single)


def requeue_single(pipeline, single):

    task = pipeline.tasks.get(single)

    if task is None:
        raise Exception(f"pipeline has no task with key '{single}'")

    task.re_queue()
    task.prepare()

    print(f"Task {single} requeued")


@click.command()
@click.pass_context
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--env', type=click.STRING, default=None)
@click.option('--instances-dir', type=click.STRING)
def watch(ctx, pipeline, env, instances_dir):

    if instances_dir is None:
        raise Exception(
            f"serve-multi requires mandatory option --instances-dir=<directory of pipeline instances>"
        )

    pipeline = _pipeline_from_modul_func(None, pipeline, write_hint_file_if_not_exists=False)

    if not os.path.exists(instances_dir):
        pathlib.Path(instances_dir).mkdir(parents=True)

    janitor = Janitor(pipeline, pipeline_instances_dir=instances_dir)

    thread = janitor.start(stay_alive_when_no_more_work=True)

    janitor.start_remote_janitors()

    thread.join()

    os._exit(0)


def _validate_task_generator_callable_with_single_dsl_arg_and_get_module_file(task_generator):

    a = inspect.signature(task_generator)

    f = os.path.dirname(os.path.abspath(inspect.getmodule(task_generator).__file__))

    if len(a.parameters) != 1:
        raise Exception(f"function given to cli.run() should take a singe DryPipe.dsl() as argument, and yield tasks" +
                        f" was given {len(a.parameters)} args")

    return f


@click.command()
@click.pass_context
@click.option('--instances-dir', type=click.STRING)
@click.option('--bind', default="0.0.0.0", help="bind address for --web-mon")
@click.option('--port', default=5000)
def serve_ui(ctx, instances_dir, bind, port):
    from dry_pipe.websocket_server import WebsocketServer

    if instances_dir is None:
        raise Exception(
            f"serve-multi requires mandatory option --instances-dir=<directory of pipeline instances>"
        )

    if not os.path.exists(instances_dir):
        pathlib.Path(instances_dir).mkdir(parents=True)

    WebsocketServer.start(bind, port, instances_dir)

    os._exit(0)


def _func_from_mod_func(mod_func):

    mod, func_name = mod_func.split(":")

    if not mod.startswith("."):
        module = importlib.import_module(mod)
    else:
        module = importlib.import_module(mod[1:])

    python_task = getattr(module, func_name, None)
    if python_task is None:
        raise Exception(f"function {func_name} not found in module {mod}")

    return python_task


@click.command()
@click.pass_context
@click.argument('mod-func', type=click.STRING)
def test(ctx, mod_func):
    python_task = _func_from_mod_func(mod_func)

    if len(python_task.tests) == 0:
        click.echo(f"No tests for python_task {mod_func}")
        return

    for test in python_task.tests:
        python_task.func(*test["args"], test=test)


@click.command()
@click.pass_context
@click.argument('mod-func', type=click.STRING)
@click.option('--task-env', type=click.Path(exists=True, dir_okay=False))
def call(ctx, mod_func, task_env):

    python_task = _func_from_mod_func(mod_func)

    env = os.environ

    if task_env is not None:
        # or if pipeline provided --task=key
        env = env_from_sourcing(task_env)

    #TODO: Validate missing args
    args = [
        env.get(k)
        for k, v in python_task.signature.parameters.items()
    ]

    out_vars = None

    try:
        out_vars = python_task.func(* args)
    except Exception as ex:
        traceback.print_exc()
        exit(1)

    # TODO: complain if missing task __X vars

    control_dir = env["__control_dir"]

    try:
        if out_vars is not None:
            with open(env["__output_var_file"], "a+") as f:
                for k, v in out_vars.items():
                    try:
                        v = json.dumps(v)
                    except TypeError as ex0:
                        print(
                            f"task call {control_dir}.{mod_func} returned var {k} of unsupported type: {type(v)}" +
                            " only primitive types are supported.",
                            file=sys.stderr
                        )
                        exit(1)
                    f.write(f"{k}={v}\n")

        for pid_file in glob.glob(os.path.join(control_dir, "*.pid")):
            os.remove(pid_file)

    except Exception as ex:
        traceback.print_exc()
        exit(1)


@click.command()
@click.pass_context
@click.option('--downstream-of', type=click.STRING)
@click.option('--not-matching', type=click.STRING)
@click.option('--matching', type=click.STRING)
@click.option('--instance-dir', type=click.Path(), default=None)
def reset(ctx, downstream_of, not_matching, matching, instance_dir):

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if matching is not None or not_matching is not None:

        if matching is not None and not_matching is not None:
            raise Exception("--matching and --not-matching can't be both specified")

        def do_reset(key):
            if matching is not None:
                return re.match(matching, key)
            return not re.match(not_matching, key)

        for task in pipeline.tasks:
            if do_reset(task.key):
                if click.confirm(f"reset {task.key} ?"):
                    task.clean()
                    print(f"{task.key} reset")

    elif downstream_of is not None:

        after_task_including = pipeline.tasks.get(downstream_of)

        if after_task_including is None:
            raise Exception(f"pipeline has no task with key '{downstream_of}'")

        def all_upstream_deps(task):
            yield task
            for upstream_task, _1, _2 in task.upstream_deps_iterator():
                for t in all_upstream_deps(upstream_task):
                    yield t

        raise Exception("not implemented")
    else:
        if click.confirm("reset all tasks ?"):
            print("boom")

@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
def stats(ctx, instance_dir):

    the_stats = None

    if instance_dir is not None:
        the_stats = fetch_task_groups_stats(instance_dir)
    else:
        pipeline_func = ctx.obj["pipeline_func"]
        pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)
        the_stats = fetch_task_groups_stats(pipeline.pipeline_instance_dir)

    for stat_row in the_stats:
        print("\t".join([str(v) for v in stat_row]))


@click.command()
@click.pass_context
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
def prepare_remote_sites(pipeline):

    pipeline_spec = pipeline

    if pipeline_spec is None:
        click.echo("must specify pipeline, ex: -p <aModule>:<aPipelineFunction>")
        return

    pipeline = _pipeline_from_modul_func(None, pipeline_spec, write_hint_file_if_not_exists=False)

    pipeline.prepare_remote_sites()


@click.command()
@click.pass_context
def status(ctx):

    raise Exception(f"not implemented")

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = DryPipe.create_pipeline(pipeline_func)
    console = Console()
    console.print(_status_table(pipeline))


def _register_commands():
    cli_group.add_command(run)
    cli_group.add_command(mon)
    cli_group.add_command(prepare)
    cli_group.add_command(call)
    cli_group.add_command(test)
    cli_group.add_command(status)
    cli_group.add_command(watch)
    cli_group.add_command(serve_ui)
    cli_group.add_command(clean)
    cli_group.add_command(stats)
    cli_group.add_command(reset)
    cli_group.add_command(requeue)

def run_cli():
    _register_commands()
    cli_group()

if __name__ == '__main__':
    run_cli()



import os
import subprocess
import time
from paramiko import SSHClient, AutoAddPolicy
import logging.config

import threading

from dry_pipe import Executor
from dry_pipe.bash import bash_shebang
from dry_pipe.internals import flatten
from dry_pipe.task_state import TaskState
from dry_pipe.utils import perf_logger_timer

logger_ssh = logging.getLogger(f"{__name__}.ssh")


class SftpFileNotFoundError(Exception):
    """
     sftp.put fragile, Paramiko bug:  https://github.com/paramiko/paramiko/issues/149
    """
    pass


class RemoteSSH(Executor):

    def __init__(
        self, ssh_username, ssh_host, remote_base_dir, key_filename, before_execute_bash
    ):
        if key_filename is None:
            raise Exception("key_filename can't be none")

        self.before_execute_bash = before_execute_bash
        self.remote_base_dir = remote_base_dir
        self.ssh_host = ssh_host
        self.ssh_username = ssh_username
        self.key_filename = key_filename
        self.rsync_containers = True
        self.slurm = None
        self.ssh_timeout_in_seconds = 60

    def is_remote(self):
        return True

    def user_at_host(self):
        return f"{self.ssh_username}@{self.ssh_host}"

    def server_connection_key(self):
        return f"{self.ssh_username}-{self.ssh_host}"

    def connect(self):
        try:
            self.ssh_client = SSHClient()
            self.ssh_client.set_missing_host_key_policy(AutoAddPolicy)

            with perf_logger_timer("RemoteSSH.connect") as t:
                self.ssh_client.connect(
                    self.ssh_host,
                    username=self.ssh_username,
                    key_filename=os.path.expanduser(self.key_filename)
                        if self.key_filename is not None and self.key_filename.startswith("~")
                        else self.key_filename,
                    timeout=self.ssh_timeout_in_seconds
                )
            logger_ssh.info("new ssh connection established %s at %s", self.ssh_username, self.ssh_host)
        except Exception as ex:
            logger_ssh.error(f"ssh connect failed: {self}")
            raise ex

    def invoke_remote(self, cmd, bash_error_ok=False):
        logger_ssh.debug("will invoke '%s' at %s", cmd, self.ssh_host)

        with perf_logger_timer("RemoteSSH.invoke_remote", cmd) as t:
            stdin, stdout, stderr = self.ssh_client.exec_command(cmd, timeout=self.ssh_timeout_in_seconds)

            stdout_txt = stdout.read().decode("utf-8")

        if not bash_error_ok and stdout.channel.recv_exit_status() != 0:
            stderr = stderr.read().decode("utf-8")
            raise Exception(f"remote call failed '{cmd}'\n{stderr}\non {self}")

        logger_ssh.debug(f"invocation of '%s' at {self} returned '%s'", cmd, stdout_txt)

        return stdout_txt

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.ssh_client.close()

    def fetch_remote_task_states(self, pipeline):
        with perf_logger_timer("RemoteSSH.fetch_remote_task_states") as t:
            remote_pid_basename = os.path.basename(pipeline.pipeline_instance_dir)

            cmd = f"find {self.remote_base_dir}/{remote_pid_basename}/.drypipe/*/state.* 2>/dev/null || true"
            stdout = self.invoke_remote(cmd)

            return [
                TaskState(f)
                for f in stdout.strip().split("\n")
                if f != ""
            ]

    def fetch_logs_and_history(self, task):

        remote_pid_basename = os.path.basename(task.pipeline_instance.pipeline_instance_dir)

        #Not called because too slow !
        def fetch_remote_state():
            cmd = f"find {self.remote_base_dir}/{remote_pid_basename}/.drypipe/{task.key}/state.* 2>/dev/null"

            stdout = self.invoke_remote(cmd)

            return TaskState(stdout.strip())

        with self.ssh_client.open_sftp() as sftp:

            def file_content_and_last_mod_time(p):
                f = os.path.join(self.remote_base_dir, remote_pid_basename, p)
                s = sftp.stat(f)
                with sftp.open(f) as _f:
                    return [_f.read(), s.st_mtime]

            f_out, t_out, \
            f_err, t_err, \
            f_history_file, t_history_file = flatten(map(file_content_and_last_mod_time, [
                task.out_log(),
                task.err_log(),
                task.history_file()
            ]))

            last_activity_time = max([t_err, t_out, t_history_file])

            return f_out.decode("utf-8"), f_err.decode("utf-8"), f_history_file.decode("utf-8"), last_activity_time

    def _remote_control_dir(self, task):
        return os.path.join(
            self.remote_base_dir,
            os.path.basename(task.pipeline_instance.pipeline_instance_dir),
            ".drypipe",
            task.key
        )

    def _file_in_control_dir(self, task, file):
        return os.path.join(
            self._remote_control_dir(task),
            file
        )

    def clear_remote_task_state(self, task):
        state_file = self._file_in_control_dir(task, "state.*")
        self.invoke_remote(f"rm {state_file}", bash_error_ok=True)

    def kill_slurm_task(self, task):

        slurm_job_id_file = self._file_in_control_dir(task, "slurm_job_id")

        with self.ssh_client.open_sftp() as sftp:

            with sftp.open(slurm_job_id_file) as _f:
                slurm_job_id = _f.read().strip().decode("utf-8")
                self.invoke_remote(f"scancel {slurm_job_id}")

    def delete_remote_state_file(self, file):

        self.invoke_remote(f"rm {file}")

    def _rsync_with_args_and_remote_dir(self):

        ssh_args = f"-e 'ssh -q -i {self.key_filename} -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null'"

        timeout = f"--timeout={60 * 2}"

        return (
            f"rsync {ssh_args} {timeout}",
            f"{self.ssh_username}@{self.ssh_host}:{self.remote_base_dir}"
        )

    def _launch_command(self, command, exception_func=lambda stderr_text: stderr_text):

        with subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            text=True
        ) as p:
            p.wait()

            if p.returncode != 0:
                raise exception_func(p.stderr.read().strip())

    def do_rsync_container(self, image_path):

        rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()

        bn = os.path.basename(image_path)
        dn = os.path.dirname(image_path)

        self.invoke_remote(f"mkdir -p {self.remote_base_dir}/pipeline_code_dir")

        cmd = f"{rsync_call} -az --partial {image_path} {remote_dir}/pipeline_code_dir"

        self._launch_command(
            cmd,
            lambda err: Exception(f"uploading container {bn} failed:\n{cmd}\n{err}")
        )

    def rsync_remote_code_dir_if_applies(self, pipeline, task_conf):

        remote_pipeline_code_dir = task_conf.remote_pipeline_code_dir

        if remote_pipeline_code_dir is None:
            return

        with perf_logger_timer("RemoteSSH.rsync_remote_code_dir_if_applies") as t:

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()

            self.invoke_remote(f"mkdir -p {remote_pipeline_code_dir}")

            cmd = f"{rsync_call} -az --partial " + \
                  f"{pipeline.pipeline_code_dir}/ {self.ssh_username}@{self.ssh_host}:{remote_pipeline_code_dir}"

            self._launch_command(
                cmd,
                lambda err: Exception(f"rsync of remote_pipeline_code_dir failed:\n{cmd}\n{err}")
            )

    def upload_task_inputs(self, task_state, task):
        with perf_logger_timer("RemoteSSH.upload_task_inputs") as t:

            task_control_dir = task_state.control_dir()

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()
            pipeline_instance_dir = os.path.dirname(os.path.dirname(task_control_dir))

            remote_pid_basename = os.path.basename(pipeline_instance_dir)

            self.invoke_remote(f"mkdir -p {self.remote_base_dir}/{remote_pid_basename}")

            cmd = f"{rsync_call} -caRz --partial --recursive --files-from={task_control_dir}/local-deps.txt " + \
                  f"{pipeline_instance_dir} {remote_dir}/{remote_pid_basename}"

            self._launch_command(
                cmd,
                lambda err: Exception(f"uploading of task inputs {task_control_dir} failed:\n{cmd}\n{err}")
            )

            # NO LONGER SUPPORTED
            # if self.rsync_containers and task.container is not None:
            #    self.do_rsync_container(task.container.image_path)

    def download_task_results(self, task_state):
        with perf_logger_timer("RemoteSSH.download_task_results") as t:
            task_control_dir = task_state.control_dir()

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()
            pipeline_instance_dir = os.path.dirname(os.path.dirname(task_control_dir))
            remote_pid_basename = os.path.basename(pipeline_instance_dir)

            cmd = f"{rsync_call} --dirs -aR --partial --files-from={task_control_dir}/remote-outputs.txt " + \
                  f"{remote_dir}/{remote_pid_basename} {pipeline_instance_dir}"

            def error_msg(err):
                r = f"{task_control_dir}/remote-outputs.txt"
                return Exception(
                    f"downloading of task results {r} from {self.ssh_host} failed:\n{cmd}\n{err}"
                )

            self._launch_command(cmd, error_msg)

    def upload_overrides(self, pipeline_instance, task_conf):

        if pipeline_instance.is_remote_overrides_uploaded(self.server_connection_key()):
            return

        pipeline_instance_dir = pipeline_instance.pipeline_instance_dir
        remote_pid_basename = os.path.basename(pipeline_instance_dir)

        overrides_file_for_host = os.path.join(
            pipeline_instance._work_dir,
            f"pipeline-env-{self.server_connection_key()}.sh"
        )

        logger_ssh.info(f"will upload overrides: '%s'", overrides_file_for_host)

        remote_pipeline_code_dir = task_conf.remote_pipeline_code_dir
        remote_containers_dir = task_conf.remote_containers_dir

        if remote_containers_dir is None:
            if remote_pipeline_code_dir is not None:
                remote_containers_dir = os.path.join(remote_pipeline_code_dir, "containers")

        def gen_remote_overrides():
            yield "__pipeline_code_dir", remote_pipeline_code_dir
            if remote_containers_dir is not None:
                yield "__containers_dir", remote_containers_dir

        remote_overrides = [
            f"export {k}={v}\n"
            for k, v in gen_remote_overrides()
            if v is not None
        ]

        if len(remote_overrides) == 0:
            logger_ssh.info(f"no overrides to upload")
            return

        with open(overrides_file_for_host, "w") as _f:
            _f.write(f"{bash_shebang()}\n\n")
            _f.writelines(remote_overrides)
            _f.write("\n")

        with perf_logger_timer("RemoteSSH.upload_overrides") as t:

            r_work_dir = os.path.join(self.remote_base_dir, remote_pid_basename, ".drypipe")
            r_override_file = os.path.join(
                r_work_dir,
                "pipeline-env.sh"
            )

            def upload_overrides():
                self.invoke_remote(f"mkdir -p {r_work_dir}")
                with self.ssh_client.open_sftp() as sftp:

                    # sftp.put fragile, Paramiko bug:  https://github.com/paramiko/paramiko/issues/149
                    for i in range(1, 3):
                        try:
                            sftp.put(overrides_file_for_host, r_override_file, confirm=True)
                            logger_ssh.info(f"overrides loaded: '%s'", r_override_file)
                            break
                        except Exception as ex:
                            if i >= 3:
                                logger_ssh.warning(ex)
                                raise SftpFileNotFoundError()
                            else:
                                logger_ssh.warning(f"sftp failure %s will sleep and retry", i)
                                time.sleep(2)

            # as a work around
            def upload_overrides_work_around():

                def lines_in_overrides_file():
                    yield f"mkdir -p {r_work_dir}"
                    yield f"echo '{bash_shebang()}' > {r_override_file}"
                    if remote_pipeline_code_dir is not None:
                        yield f'echo "export __pipeline_code_dir={remote_pipeline_code_dir}" >> {r_override_file}'
                    if remote_containers_dir is not None:
                        yield f'echo "export __containers_dir={remote_containers_dir}" >> {r_override_file}'

                    yield f"chmod u+x {r_override_file}"

                for i in range(1, 3):
                    try:
                        self.invoke_remote("pwd")
                        break
                    except Exception as ex:
                        if i >= 3:
                            raise ex
                        else:
                            time.sleep(2)

                remote_cmd = " && ".join(lines_in_overrides_file())
                self.invoke_remote(remote_cmd)

            #upload_overrides_work_around()

            upload_overrides()

            pipeline_instance.set_remote_overrides_uploaded(self.server_connection_key())

    """
        Fetches log lines and history.txt for all tasks, done once every janitor run, instead of before each task
    """
    def fetch_new_log_lines(self, pipeline_instance_dir):
        with perf_logger_timer("RemoteSSH.fetch_new_log_lines") as t:

            rsync_call, remote_dir = self._rsync_with_args_and_remote_dir()
            remote_pid_basename = os.path.basename(pipeline_instance_dir)
            dst = os.path.dirname(pipeline_instance_dir)

            cmd = " ".join([
                f" {rsync_call} -r --partial --append",
                " --filter='+ .drypipe/*/*.log'",
                " --filter='+ .drypipe/*/history.tsv'",
                " --filter='- .drypipe/*/*'",
                " --filter='- publish'",
                " --filter='- *.sh'",
                f" {remote_dir}/{remote_pid_basename} {dst}"
            ])

            logger_ssh.debug("will rsync new log lines \n%s", cmd)

            self._launch_command(
                cmd,
                lambda err: Exception(
                    f"fetching logs failed:\n{cmd}\n{err}")
            )

    def execute(self, task, touch_pid_file_func, wait_for_completion=False, fail_silently=False):
        with perf_logger_timer("RemoteSSH.execute") as t:
            b4_command = ""
            if self.before_execute_bash is not None:
                b4_command = f"{self.before_execute_bash} &&"

            remote_pid_basename = os.path.basename(task.pipeline_instance.pipeline_instance_dir)

            def remote_base_dir(p):
                return os.path.join(self.remote_base_dir, remote_pid_basename, p)

            r_script, r_sbatch_script, r_out, r_err, r_work_dir, r_control_dir = map(remote_base_dir, [
                task.script_file(),
                task.sbatch_launch_script(),
                task.out_log(),
                task.err_log(),
                task.work_dir(),
                task.control_dir()
            ])

            self.invoke_remote(f"mkdir -p {r_work_dir}")
            self.invoke_remote(f"mkdir -p {r_control_dir}/out_sigs")

            if task.scratch_dir() is not None:
                self.invoke_remote(f"mkdir -p {remote_base_dir(task.scratch_dir())}")

            self.invoke_remote(f"rm {r_control_dir}/state.*", bash_error_ok=True)

            self.invoke_remote(f"touch {r_control_dir}/state.launched.0")

            if self.slurm is None:

                ampersand_or_not = "&"

                if wait_for_completion:
                    ampersand_or_not = ""

                cmd = f"nohup bash -c '{b4_command} . {r_script}  >{r_out} 2>{r_err}' {ampersand_or_not}"
            else:
                if wait_for_completion:
                    cmd = f"bash -c 'export SBATCH_WAIT=True && {r_sbatch_script}'"
                else:
                    cmd = r_sbatch_script

            self.invoke_remote(cmd)



from setuptools import setup

minimal_requirements = [
    'click==8.1.3',
    'uvicorn==0.17.6',
    'python-socketio==5.6.0',
    'psutil==5.9.1',
    'rich==12.4.4',
    'requests==2.27.1',
    'sshkeyboard==2.3.1'
]

extra_requirements = [
    'PyYAML==5.4.1',
    'paramiko==2.11.0'
]

setup(
    name='dry_pipe',
    version='0.2.18',
    packages=['dry_pipe'],
    install_requires=minimal_requirements,
    extras_require={"full": extra_requirements},
    include_package_data=True,
    entry_points='''
        [console_scripts]
        drypipe=dry_pipe.cli:run_cli
    ''',
)